import logging

import boto3
from airflow.sensors.base_sensor_operator import BaseSensorOperator
from airflow.utils.decorators import apply_defaults

logger = logging.getLogger(None)


class MonitorGlueJob(BaseSensorOperator):
    """
    Retrieves the metadata for a given job run and checks the run status.
    """

    template_fields = ("poke_interval", "timeout", "soft_fail", "mode", "xcom_tasks")

    @apply_defaults
    def __init__(
        self,
        poke_interval=60,
        timeout=60 * 60 * 24 * 7,
        soft_fail=False,
        mode="poke",
        xcom_tasks=None,
        *args,
        **kwargs,
    ):
        super(MonitorGlueJob, self).__init__(
            poke_interval=poke_interval,
            timeout=timeout,
            soft_fail=soft_fail,
            mode=mode,
            *args,
            **kwargs,
        )
        self.xcom_tasks = xcom_tasks

    def poke(self, context):
        """Monitors Glue Job Run Status"""

        CLIENT = boto3.client("glue")

        task_ids = self.xcom_tasks["job_name"]["task_id"]

        job_name = context["task_instance"].xcom_pull(
            task_ids=task_ids, key=self.xcom_tasks["job_name"]["key"]
        )

        job_run_id = context["task_instance"].xcom_pull(
            task_ids=task_ids, key=self.xcom_tasks["job_run_id"]["key"]
        )

        try:
            response = CLIENT.get_job_run(
                JobName=job_name,
                RunId=job_run_id,
            )

            state = response["JobRun"]["JobRunState"]
            logging.info(f"Current state for {job_name} Glue job : {state}")
            if state == "SUCCEEDED":
                # success
                return True
            elif state in ["STARTING", "RUNNING", "STOPPING"]:
                # retry state
                return False
            else:
                # failure, up for retry
                raise Exception(
                    "training job status == {JOB_STATE} with reason = {REASON}".format(
                        JOB_STATE=state, REASON=response["JobRun"]["ErrorMessage"]
                    )
                )

        except Exception as err:
            raise


class MonitorGlueCrawler(BaseSensorOperator):
    """
    Retrieves the metadata for a given crawler run and checks the run status.
    """

    template_fields = ("poke_interval", "timeout", "soft_fail", "mode", "xcom_tasks")

    @apply_defaults
    def __init__(
        self,
        poke_interval=60,
        timeout=60 * 60 * 24 * 7,
        soft_fail=False,
        mode="poke",
        xcom_tasks=None,
        *args,
        **kwargs,
    ):
        super(MonitorGlueCrawler, self).__init__(
            poke_interval=poke_interval,
            timeout=timeout,
            soft_fail=soft_fail,
            mode=mode,
            *args,
            **kwargs,
        )
        self.xcom_tasks = xcom_tasks

    def poke(self, context):
        """Monitors Glue Crawler Run Status"""

        CLIENT = boto3.client("glue")

        task_ids = self.xcom_tasks["crawler_name"]["task_id"]

        crawler_name = context["task_instance"].xcom_pull(
            task_ids=task_ids, key=self.xcom_tasks["crawler_name"]["key"]
        )

        try:
            response = CLIENT.get_crawler(Name=crawler_name)

            state = response["Crawler"]["State"]
            if state == "READY":
                # success
                return True
            elif state in ["RUNNING", "STOPPING"]:
                # retry state
                return False
            else:
                # failure, up for retry
                raise Exception(
                    "training job status == {JOB_STATE} with reason = {REASON}".format(
                        JOB_STATE=state, REASON=state
                    )
                )

        except Exception as err:
            raise



# pylint: disable=super-with-arguments
# pylint: disable=redefined-builtin
# pylint: disable=invalid-name

import logging

import boto3
from airflow.sensors.base_sensor_operator import BaseSensorOperator
from airflow.utils.decorators import apply_defaults

logger = logging.getLogger(None)


class MonitorEcsTask(BaseSensorOperator):
    """
    Retrieves the metadata for a given job run and checks the run status.
    """

    template_fields = ("config", "xcom_tasks")

    @apply_defaults
    def __init__(self, config, xcom_tasks=None, *args, **kwargs):
        super(MonitorEcsTask, self).__init__(*args, **kwargs)

        self.config = config
        self.xcom_tasks = xcom_tasks

    def poke(self, context):
        """Monitors ECS Task Run Status"""

        CLIENT = boto3.client("ecs")

        task_cluster = self.config["task_cluster"]
        task_ids = self.xcom_tasks["task_run_id"]["task_id"]
        task_run_id = context["task_instance"].xcom_pull(
            task_ids=task_ids, key=self.xcom_tasks["task_run_id"]["key"]
        )

        try:
            response = CLIENT.describe_tasks(cluster=task_cluster, tasks=[task_run_id])
            ecs_task_definition_status = (
                response.get("tasks")[0].get("containers")[0].get("lastStatus")
            )
            logging.info(
                "Looking for ECS Task ID: {id} in Cluster: {cluster}".format(
                    id=task_run_id, cluster=task_cluster
                )
            )

            if ecs_task_definition_status in [
                "DEACTIVATING",
                "STOPPING",
                "DEPROVISIONING",
                "STOPPED",
            ]:
                return True
            elif ecs_task_definition_status in [
                "PROVISIONING",
                "PENDING",
                "ACTIVATING",
                "RUNNING",
            ]:
                return False
            else:
                raise Exception(
                    "training job status == {TASK_STATE} with reason = {REASON}".format(
                        TASK_STATE=ecs_task_definition_status, REASON=response
                    )
                )

        except Exception as err:
            raise err



# pylint: disable=super-with-arguments
# pylint: disable=redefined-builtin
# pylint: disable=invalid-name

"""
    GLUE TASK CODE
        - these are used in pythonOperator steps as a way \
            to flexibly call boto3 and perform any special requirements
"""
import logging
import time
import warnings

import boto3
from airflow.models.baseoperator import BaseOperator
from airflow.utils.decorators import apply_defaults

logger = logging.getLogger(None)


class RunGlueJob(BaseOperator):
    """Runs Glue Job, and can consume variables for the job."""

    template_fields = ("config", "xcom_tasks")

    @apply_defaults
    def __init__(self, config, xcom_tasks=None, *args, **kwargs):
        super(RunGlueJob, self).__init__(*args, **kwargs)

        self.config = config
        self.xcom_tasks = xcom_tasks

    def execute(self, context):
        """
        :config.job_name: str. the name of the glue job to run.
        :config.job_arguments: dict. the key value arguments send to the job.
        :xcom_push: str. glue_job_run_id.
        """
        CLIENT = boto3.client("glue")

        job_name = self.config.get("job_name", None)
        job_arguments = self.config.get("job_arguments", None)

        if job_name is None:
            raise ValueError("Job name is required for task.")
        if job_arguments is None:
            warnings.warn("Job arguments are empty.")

        try:
            logging.info(
                "Running glue job {name} with envs {envs}.".format(
                    name=job_name, envs=job_arguments
                )
            )

            response = CLIENT.start_job_run(JobName=job_name, Arguments=job_arguments)

            # push job_run_id to xcom
            context["task_instance"].xcom_push(
                key="job_run_id", value=response["JobRunId"]
            )

            # push job_name to xcom for other glue tasks to reference
            context["task_instance"].xcom_push(key="job_name", value=job_name)

        except Exception as err:
            raise err


class TagGlueJob(BaseOperator):
    """Adds a tag to the resource specified."""

    template_fields = (
        "config",
        "xcom_tasks",
    )

    @apply_defaults
    def __init__(self, config, xcom_tasks=None, *args, **kwargs):
        super(TagGlueJob, self).__init__(*args, **kwargs)

        self.config = config
        self.xcom_tasks = xcom_tasks

    def execute(self, context):
        """ """
        CLIENT = boto3.client("glue")

        job_name = self.config.get("job_name", None)
        resource_tags = self.config.get("resource_tags", None)

        if job_name is None:
            # if optional config job_name is None, pull from xcom
            job_name = context["task_instance"].xcom_pull(
                self.xcom_tasks["job_name"]["task_id"],
                key=self.xcom_tasks["job_name"]["key"],
            )

        if resource_tags is None:
            raise ValueError("Resource tags are required for task.")

        try:
            account_id = boto3.client("sts").get_caller_identity().get("Account")
            glue_resource_arn = "arn:aws:glue:eu-west-1:{account_id}:job/{name}".format(
                account_id=account_id, name=job_name
            )

            CLIENT.tag_resource(ResourceArn=glue_resource_arn, TagsToAdd=resource_tags)

            # push resource_tags to xcom for other glue tasks to reference
            context["task_instance"].xcom_push(key="resource_tags", value=resource_tags)

        except Exception as err:
            raise err


class UnTagGlueJob(BaseOperator):
    """Removes a tag to the resource specified."""

    template_fields = (
        "config",
        "xcom_tasks",
    )

    @apply_defaults
    def __init__(self, config, xcom_tasks=None, *args, **kwargs):
        super(UnTagGlueJob, self).__init__(*args, **kwargs)

        self.config = config
        self.xcom_tasks = xcom_tasks

    def execute(self, context):
        """
        :glue_resource_arn: str. the Arn of the glue resource.
        :resource_tags: dict. the key value pairs of the
                             relevant tags to remove.
        """
        CLIENT = boto3.client("glue")

        job_name = self.config.get("job_name", None)
        resource_tags = self.config.get("resource_tags", None)

        task_ids = self.xcom_tasks["job_name"]["task_id"]

        if job_name is None:
            # if optional config job_name is None, pull from xcom
            job_name = context["task_instance"].xcom_pull(
                task_ids=task_ids, key=self.xcom_tasks["job_name"]["key"]
            )

        if resource_tags is None:
            # if optional config resource_tags None, pull from xcom
            resource_tags = context["task_instance"].xcom_pull(
                task_ids=task_ids, key=self.xcom_tasks["resource_tags"]["key"]
            )

        try:
            account_id = boto3.client("sts").get_caller_identity().get("Account")
            glue_resource_arn = "arn:aws:glue:eu-west-1:{account_id}:job/{name}".format(
                account_id=account_id, name=job_name
            )

            tags_to_remove = list(key for key in resource_tags.keys())

            CLIENT.untag_resource(
                ResourceArn=glue_resource_arn, TagsToRemove=tags_to_remove
            )

        except Exception as err:
            raise err


class RunGlueCrawler(BaseOperator):
    """Runs Glue Crawler."""

    template_fields = (
        "config",
        "xcom_tasks",
    )

    @apply_defaults
    def __init__(self, config=None, xcom_tasks=None, *args, **kwargs):
        super(RunGlueCrawler, self).__init__(*args, **kwargs)

        self.config = config
        self.xcom_tasks = xcom_tasks

    def execute(self, context):
        """
        :glue_crawler_name: str. the name of the crawler to be triggered.
        """
        CLIENT = boto3.client("glue")

        crawler_name = self.config.get("crawler_name", None)
        if crawler_name is None:
            raise ValueError("Crawler name needed for task.")

        try:
            logging.info("Running glue crawler {name}".format(name=crawler_name))

            CLIENT.start_crawler(Name=crawler_name)

            # push crawler_name to xcom
            context["task_instance"].xcom_push(key="crawler_name", value=crawler_name)

        except Exception as err:
            raise err



# pylint: disable=super-with-arguments
# pylint: disable=redefined-builtin
# pylint: disable=invalid-name

import logging
import re

import boto3
from airflow.sensors.base_sensor_operator import BaseSensorOperator
from airflow.utils.decorators import apply_defaults

logger = logging.getLogger(None)


class MonitorEcsTask(BaseSensorOperator):
    """
    Retrieves the metadata for a given job run and checks the run status.
    """

    template_fields = (
        "poke_interval",
        "timeout",
        "soft_fail",
        "mode",
        "xcom_tasks",
        "config",
    )

    @apply_defaults
    def __init__(
        self,
        poke_interval=60,
        timeout=60 * 60 * 24 * 7,
        soft_fail=False,
        mode="poke",
        xcom_tasks=None,
        config=None,
        *args,
        **kwargs,
    ):
        super(MonitorEcsTask, self).__init__(
            poke_interval=poke_interval,
            timeout=timeout,
            soft_fail=soft_fail,
            mode=mode,
            *args,
            **kwargs,
        )

        self.config = config
        self.xcom_tasks = xcom_tasks

    def poke(self, context):
        CLIENT = boto3.client("ecs")

        task_cluster = self.config["task_cluster"]
        task_ids = self.xcom_tasks["task_run_id"]["task_id"]
        task_run_id = context["task_instance"].xcom_pull(
            task_ids=task_ids, key=self.xcom_tasks["task_run_id"]["key"]
        )
        response = CLIENT.describe_tasks(cluster=task_cluster, tasks=[task_run_id])
        logging.info(f"Looking for ECS Task: {task_run_id} in Cluster: {task_cluster}")
        try:
            if len(response.get("failures", [])) > 0:
                raise Exception(response.get("failures"))

            task = response["tasks"][0]
            task_stop_reason = task.get("stoppedReason", "")
            if task.get("stopCode", "") == "TaskFailedToStart":
                raise Exception(f"The task failed to start due to: {task_stop_reason}")

            #     # This is a `stoppedReason` that indicates a task has not successfully
            #     # finished, but there is no other indication of failure in the response.
            #     # https://docs.aws.amazon.com/AmazonECS/latest/developerguide/stopped-task-errors.html
            if re.match(
                r"Host EC2 \(instance .+?\) (stopped|terminated)\.", task_stop_reason
            ):
                raise Exception(
                    f"The task was stopped because the host instance terminated: {task_stop_reason}"
                )
            containers = task["containers"]
            running_statuses = [
                "PROVISIONING",
                "PENDING",
                "ACTIVATING",
                "RUNNING",
                "DEACTIVATING",
                "STOPPING",
                "DEPROVISIONING",
            ]
            for container in containers:
                container_name = container.get("name")
                container_last_status = container.get("lastStatus")
                container_reason = container.get("reason", "").lower()
                logging.info(f"Last Status : {container_last_status}")
                if container_last_status == "STOPPED":
                    if container["exitCode"] != 0:
                        raise Exception(
                            f"Container did not exit successfully {container_name}"
                        )
                    logging.info(f"Container exited successfully : {container_name}")
                    return True
                elif container_last_status in running_statuses:
                    return False
                elif "error" in container_reason:
                    raise Exception(
                        f"Container encountered an error when launching :{container_name} : {container_reason}"
                    )

        except Exception as err:
            raise err



# pylint: disable=super-with-arguments
# pylint: disable=redefined-builtin
# pylint: disable=invalid-name

"""SDC Custom S3 Operators"""
"""SDC Custom S3 Operators"""
import os
import logging
from urllib.parse import urlparse
import boto3
from airflow.models.baseoperator import BaseOperator
from airflow.utils.decorators import apply_defaults

logger = logging.getLogger(None)


def parse_s3_url(url):
    """Returns an (s3 bucket, key name/prefix) tuple from a url with an s3 scheme.
    Args:
        url (str):
    Returns:
        tuple: A tuple containing:
            - str: S3 bucket name
            - str: S3 key
    """
    parsed_url = urlparse(url)
    if parsed_url.scheme != "s3":
        raise ValueError(
            "Expecting 's3' scheme, got: {} in {}.".format(parsed_url.scheme, url)
        )
    return parsed_url.netloc, parsed_url.path.lstrip("/")


class CopyBetweenS3Locations(BaseOperator):
    template_fields = ("dst_path", "src_path", "keep_filename", "xcom_tasks")

    @apply_defaults
    def __init__(
        self,
        dst_path,
        src_path=None,
        keep_filename=False,
        xcom_tasks=None,
        *args,
        **kwargs
    ):
        super(CopyBetweenS3Locations, self).__init__(*args, **kwargs)
        self.dst_path = dst_path
        self.src_path = src_path
        self.xcom_tasks = xcom_tasks
        self.keep_filename = keep_filename

    def execute(self, context):
        """Copy s3 object between two buckets"""
        # init boto session and sagemaker
        BOTO_SESSION = boto3.Session()
        SM_RESOURCE = BOTO_SESSION.resource("s3")

        try:
            if self.src_path is None:
                self.src_path = context["task_instance"].xcom_pull(
                    task_ids=self.xcom_tasks["src_path"]["task_id"],
                    key=self.xcom_tasks["src_path"]["key"],
                )
                assert (
                    self.src_path is not None
                ), "No source path found. Either hardcode in config or pass from previous job"

            src_bucket, src_filepath = parse_s3_url(self.src_path)
            dst_bucket, dst_filepath = parse_s3_url(self.dst_path)

            if self.keep_filename:
                # assumes destination path is a folder path
                # keeps src filename
                src_filename = src_filepath.split("/")[-1]
                logger.info("Keep source filename = {}".format(src_filename))
                # assumes dst_path given without filename
                dst_filepath = os.path.join(dst_filepath, src_filename)
                # assumes dst_path is given without filename
                self.dst_path = os.path.join(self.dst_path, src_filename)

            copy_source = {"Bucket": src_bucket, "Key": src_filepath}
            bucket = SM_RESOURCE.Bucket(dst_bucket)
            bucket.copy(copy_source, dst_filepath)

            # push response up
            context["task_instance"].xcom_push(
                key="destination_location", value=self.dst_path
            )

        except Exception as exception:
            raise exception



# pylint: disable=super-with-arguments
# pylint: disable=redefined-builtin
# pylint: disable=invalid-name

"""
    ECS TASK CODE
        - these are used in pythonOperator steps as a way \
            to flexibly call boto3 and perform any special requirements
"""
import logging
import time

import boto3
from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults

logger = logging.getLogger(None)


class RunEcsTask(BaseOperator):
    """Run ECS Task Definition."""

    template_fields = (
        "config",
        "xcom_tasks",
    )

    @apply_defaults
    def __init__(self, config=None, xcom_tasks=None, *args, **kwargs):
        super(RunEcsTask, self).__init__(*args, **kwargs)

        self.config = config
        self.xcom_tasks = xcom_tasks

    def execute(self, context):
        """
        Triggers ECS Task Definition
        """
        CLIENT = boto3.client("ecs")

        task_definition_name = self.config.get("task_definition_name")
        task_cluster = self.config.get("task_cluster")
        security_groups = self.config.get("security_groups")
        subnets = self.config.get("subnets")

        try:
            log_message = """
                Running task: {name} with configurations:
                ecs_task_cluster: {cluster}
                security_groups: {security_groups}
                subnets: {subnets}
                """.format(
                name=task_definition_name,
                cluster=task_cluster,
                security_groups=security_groups,
                subnets=subnets,
            )
            logging.info(log_message)

            response = CLIENT.run_task(
                cluster=task_cluster,
                group=task_definition_name,
                launchType="FARGATE",
                networkConfiguration={
                    "awsvpcConfiguration": {
                        "subnets": subnets,
                        "securityGroups": security_groups,
                        "assignPublicIp": "ENABLED",
                    }
                },
                propagateTags="TASK_DEFINITION",
                taskDefinition=task_definition_name,
            )

            logging.info(response)

            ecs_task_run_id = response["tasks"][0]["containers"][0]["taskArn"].split(
                "/"
            )[-1]
            context["task_instance"].xcom_push(key="task_run_id", value=ecs_task_run_id)

        except Exception as err:
            raise err
